/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_OTHR_DEBUG			FALSE
#define FE_OTHR_THREAD_VERBOSE	FALSE

#include "operator/operator.pmh"

using namespace fe;

OperatorThreaded::~OperatorThreaded(void)
{
	finish();
}

void OperatorThreaded::finish(void)
{
	Range stop;
	for(U32 m=0;m<m_threadsAlive;m++)
	{
		m_spGang->post(stop);
	}
	if(m_spGang->workers())
	{
		m_spGang->finish();
	}
	m_threadsAlive=0;
}

void OperatorThreaded::initialize(void)
{
	catalog<I32>("Threads")=2;
	catalog<I32>("Threads","min")=1;
	catalog<I32>("Threads","high")=16;

	catalog<bool>("StayAlive")=false;

	m_spGang=new Gang<Worker,Range>();
}

void OperatorThreaded::handle(Record& a_rSignal)
{
#if FE_OTHR_DEBUG
	feLog("OperatorThreaded::handle\n");
#endif

	const U32 threads=catalog<I32>("Threads");
	const bool stayAlive=catalog<bool>("StayAlive");

	if(m_threadsAlive &&
			(!stayAlive || (stayAlive && m_threadsAlive!=threads)))
	{
#if FE_OTHR_THREAD_VERBOSE
		feLog("Pre-finish stayAlive %d threadsAlive %d\n",
				stayAlive,m_threadsAlive);
#endif
		finish();
	}

	const U32 count=m_fullRange.m_end-m_fullRange.m_start+1;

	U32 jobs=threads*4;	//* TODO param

	Range range;
	for(U32 m=0;m<jobs;m++)
	{
		range.m_start=count*m/Real(jobs);
		range.m_end=count*(m+1)/Real(jobs)-1;

#if FE_OTHR_THREAD_VERBOSE
		feLog("Post [%d,%d]\n",range.m_start,range.m_end);
#endif
		m_spGang->post(range);
	}

	if(!stayAlive || !m_threadsAlive)
	{
#if FE_OTHR_THREAD_VERBOSE
		feLog("Start %d threads\n",threads);
#endif
		m_spGang->start(threads,sp<OperatorThreaded>(this));
		if(stayAlive)
		{
			m_threadsAlive=threads;
		}
	}

	if(!stayAlive)
	{
#if FE_OTHR_THREAD_VERBOSE
		feLog("Post-finish stayAlive %d threadsAlive %d\n",
				stayAlive,m_threadsAlive);
#endif
		finish();
	}

#if FE_OTHR_DEBUG
	feLog("OperatorThreaded::handle done\n");
#endif
}

void OperatorThreaded::Worker::operator()(void)
{
#if FE_OTHR_THREAD_VERBOSE
	feLog("Starting thread %d\n",m_id);
#endif
	const BWORD& wait=m_spOperatorThreaded->m_threadsAlive;
	BWORD done=FALSE;
	while(!done)
	{
		Range range;

		if(wait)
		{
#if FE_OTHR_THREAD_VERBOSE
			feLog("Thread %d waiting\n",m_id);
#endif
			m_spJobQueue->waitForJob(range);
		}
		else
		{
			done=!m_spJobQueue->take(range);
		}

#if FE_OTHR_THREAD_VERBOSE
		feLog("Thread %d Job [%d,%d]\n",m_id,range.m_start,range.m_end);
#endif

		if(range.m_start<0)
		{
			done=TRUE;
		}
		else if(!done)
		{
			m_spOperatorThreaded->run(m_id,range);
		}
	}
#if FE_OTHR_THREAD_VERBOSE
	feLog("Sync thread %d\n",m_id);
#endif
	U32 leader=m_spJobQueue->synchronize();
#if FE_OTHR_THREAD_VERBOSE
	feLog("Sync'd thread %d leader=%d\n",m_id,leader);
#endif
}
