/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __grass_Blade__
#define __grass_Blade__

namespace fe
{

/**************************************************************************//**
	@brief minimal semi-implicit curve sim

	@ingroup grass
*//***************************************************************************/
class FE_DL_EXPORT Blade
{
	public:

						Blade(void);
virtual					~Blade(void)										{}

		void			setTime(Real a_t)		{ m_t=a_t; }
		Real			time(void) const		{ return m_t; }

		void			setTimeStep(Real a_h)	{ m_h=a_h; }
		Real			timeStep(void) const	{ return m_h; }

		void			setTension(Real a_k)	{ m_k=a_k; }
		Real			tension(void) const		{ return m_k; }

		void			setRestoration(Real a_restoration)
						{	m_restoration=a_restoration; }
		Real			restoration(void) const
						{	return m_restoration; }

		void			setDrag(Real a_c)		{ m_c=a_c; }
		Real			drag(void) const		{ return m_c; }

		void			setBend(Real a_bend)	{ m_bend=a_bend; }
		Real			bend(void) const		{ return m_bend; }

		void			setMass(Real a_m)		{ m_m=a_m; }
		Real			mass(void) const		{ return m_m; }

		void			setWind(Real a_wind)
								{ m_wind=a_wind; }
		Real			wind(void) const
								{ return m_wind; }

		void			setTurbulence(Real a_turbulence)
								{ m_turbulence=a_turbulence; }
		Real			turbulence(void) const
								{ return m_turbulence; }

		void			setCenter(SpatialVector a_center)
								{ m_center=a_center; }
		SpatialVector	center(void) const
								{ return m_center; }

		void			setRadius(Real a_radius)
								{ m_radius=a_radius; }
		Real			radius(void) const
								{ return m_radius; }

		void			storeAsRest(U32 a_index);

		void			populate(void);
		void			step(U32 a_index);

		std::vector<SpatialVector>&	tangent(void)	{ return m_tangent; }
		std::vector<SpatialVector>&	location(void)	{ return m_location; }
		std::vector<Real>&	velocity(void)			{ return m_velocity; }

	private:
	class Contact
	{
		public:
							Contact(void)	{ clear(); }
							~Contact(void)									{}

			void			clear(void)
							{
								set(m_offset);
								m_active=FALSE;
							}
			void			set(SpatialVector a_offset)
							{
								m_offset=a_offset;
								m_active=TRUE;
							}

			BWORD			active(void)	{ return m_active; }
			SpatialVector	offset(void)	{ return m_offset; }

		private:
			SpatialVector	m_offset;
			BWORD			m_active;
	};

		void			computeAngles(U32 a_index, Vector<4,Real>& a_rAngle,
								const SpatialVector* a_pDelta);

		std::vector<Contact>			m_contact;
		std::vector<SpatialVector>		m_tangent;	// root direction
		std::vector<SpatialVector>		m_location;	// point locations
		std::vector<Real>				m_velocity;	// angular velocities
		std::vector< Vector<4,Real> >	m_length;	// segment length
		std::vector< Vector<4,Real> >	m_rest;		// rest angle

		Real				m_t;
		Real				m_k;
		Real				m_restoration;
		Real				m_c;
		Real				m_bend;
		Real				m_m;
		Real				m_h;
		Real				m_wind;
		Real				m_turbulence;

		SpatialVector		m_center;
		Real				m_radius;

		Matrix<4,4,Real>	m_invA;
		Matrix<4,4,Real>	m_dfdx;
		Matrix<4,4,Real>	m_hdfdx;
};

} /* namespace */

#endif /* __grass_Blade__ */

